import os
import logging
from pulp import *
from collections import defaultdict, deque

# Konfiguriere Logging
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')

def count_initial_crossings(edges):
    crossings = 0
    for index, (i, j) in enumerate(edges):
        for k, l in edges[index+1:]:
            if i < k and j > l:
                crossings += 1
    return crossings

def count_crossings_via_variables(c_vars):
    crossings = 0
    for c_var in c_vars.values():
        if c_var.varValue == 1:
            crossings += 1
    return crossings

def insert_node_in_sorted_list(node, sorted_b, y):
    # Finde die korrekte Position des Knotens basierend auf den y-Werten
    pos = 0
    while pos < len(sorted_b):
        all_less = all(y.get((min(node, other), max(node, other)), 0).varValue == 1 for other in sorted_b[pos:])
        if all_less:
            break
        pos += 1
    sorted_b.insert(pos, node)

def kahn_topological_sort_with_cycles(nodes, graph, in_degree, y):
    sorted_b = []
    zero_in_degree_queue = deque([i for i in nodes if in_degree[i] == 0])

    while zero_in_degree_queue:
        node = zero_in_degree_queue.popleft()
        sorted_b.append(node)
        for neighbor in graph[node]:
            in_degree[neighbor] -= 1
            if in_degree[neighbor] == 0:
                zero_in_degree_queue.append(neighbor)

    unsorted_nodes = set(nodes) - set(sorted_b)
    if unsorted_nodes:
        logging.warning(f"Zyklus entdeckt. Unsortierte Knoten: {unsorted_nodes}")
        # Füge unsortierte Knoten basierend auf y-Werten ein
        for node in unsorted_nodes:
            insert_node_in_sorted_list(node, sorted_b, y)

    return sorted_b

def solve_bipartite_minimization(graph_file):
    logging.info(f"Prozess für {graph_file} gestartet")
    base_name = os.path.basename(graph_file)
    new_base_name = base_name.replace('.gr', '.sol')
    output_file = os.path.join('solution_instances', new_base_name)
    logging.info(f"Die Ausgabedatei wird {output_file} sein")

    edges = []
    with open(graph_file, "r") as file:
        for line in file:
            if line.startswith('c'):
                continue
            elif line.startswith('p'):
                parts = line.split()
                n0 = int(parts[2])
                n1 = int(parts[3])
                logging.info(f"Größen der Partitionen: A={n0}, B={n1}")
            else:
                x, y = map(int, line.split())
                edges.append((x, y))
    logging.info(f"{len(edges)} Kanten geladen.")

    prob = LpProblem("Minimize_Crossings", LpMinimize)

    # Initialisierung der y-Variablen nur für i < j
    y = {(i, j): LpVariable(f"y_{i}_{j}", 0, 1, cat='Binary') 
         for i in range(n0 + 1, n0 + n1 + 1) for j in range(n0 + 1, n0 + n1 + 1) if i < j}
    c = {(i, j, k, l): LpVariable(f"c_{i}_{j}_{k}_{l}", 0, 1, cat='Binary') 
         for (i, j) in edges for (k, l) in edges if i < j and k < l and i < k and j != l}
    logging.info("y und c Variablen initialisiert.")

    # Zielfunktion
    prob += lpSum(c.values()), "Minimize total crossings"

    # Constraints für Kreuzungen
    for (i, j) in edges:
        for (k, l) in edges:
            if k > i:  # Prüfung nur, wenn k > i nach Sortierung
                if (i, j, k, l) not in c:
                    c[(i, j, k, l)] = LpVariable(f"c_{i}_{j}_{k}_{l}", 0, 1, cat='Binary')
                if j > l:
                    prob += c[(i, j, k, l)] == y[(l, j)]
                elif l > j:
                    prob += c[(i, j, k, l)] == 1 - y[(j, l)]
    """
    # Transitivitätsconstraints
    for i in range(n0 + 1, n0 + n1 - 1):
        for j in range(i + 1, n0 + n1):
            for k in range(j + 1, n0 + n1 + 1):
                prob += y[(i, k)] >= y[(i, j)] + y[(j, k)] - 1
    """
    initial_crossings = count_initial_crossings(edges)
    logging.info(f"Initial crossings edges: {initial_crossings}")

    prob.writeLP("prob.lp")
    prob.solve()
    logging.info(f"Status der Lösung: {LpStatus[prob.status]}")

    final_crossings1 = count_crossings_via_variables(c)
    logging.info(f"Final crossings VAR (after solving): {final_crossings1}")
    """
    for v in prob.variables():
        print(v.name, "=", v.varValue)
    """
    if prob.status == LpStatusOptimal:
        logging.info("Optimale Lösung gefunden. Ergebnisse werden gespeichert. Optimal arrangement of B:")
        
        graph = defaultdict(list)
        in_degree = defaultdict(int)
        
        nodes = list(range(n0 + 1, n0 + n1 + 1))
        
        # Baue das gerichtete Graph basierend auf y-Werten
        for i in nodes:
            for j in nodes:
                if i < j:
                    y_ij = y.get((i, j))
                    if y_ij is not None and y_ij.varValue is not None:
                        if y_ij.varValue == 1:
                            graph[i].append(j)
                            in_degree[j] += 1
                        elif y_ij.varValue == 0:
                            graph[j].append(i)
                            in_degree[i] += 1

        def topological_sort(nodes, graph, in_degree):
            sorted_b = []
            zero_in_degree_queue = deque([i for i in nodes if in_degree[i] == 0])

            while zero_in_degree_queue:
                node = zero_in_degree_queue.popleft()
                sorted_b.append(node)
                for neighbor in graph[node]:
                    in_degree[neighbor] -= 1
                    if in_degree[neighbor] == 0:
                        zero_in_degree_queue.append(neighbor)

            return sorted_b, set(nodes) - set(sorted_b)
        
        sorted_b = kahn_topological_sort_with_cycles(nodes, graph, in_degree, y)
        logging.info("Topolgoische Sortierung mit Zykeln erfolgreich.")
        # Füge unsortierte Knoten basierend auf y-Werten ein
        unsorted_nodes = set(nodes) - set(sorted_b)
        for node in unsorted_nodes:
            insert_node_in_sorted_list(node, sorted_b, y)

        a = count_crossings_via_variables(c)
        
        os.makedirs(os.path.dirname(output_file), exist_ok=True)
        with open(output_file, 'w') as f:
            for b in sorted_b:
                f.write(f"{b}\n")
                print(f"{b}")
        print("Crossings: ", a)
        logging.info(f"Ergebnisse in {output_file} gespeichert")
    else:
        logging.warning("Keine optimale Lösung gefunden.")

def process_directory(directory_path):
    for filename in os.listdir(directory_path):
        if filename.endswith('.gr'):
            file_path = os.path.join(directory_path, filename)
            solve_bipartite_minimization(file_path)
            logging.info(f"Verarbeitung abgeschlossen für {file_path}")

directory_path = 'githubtests/medium_test_set/instances/'
directory_path = 'githubtests/tiny_test_set/instances/'
#directory_path = 'mytests/instances/'
process_directory(directory_path)

#test_file = 'githubtests/medium_test_set/instances/10.gr'
#solve_bipartite_minimization(test_file)
