import os
import logging
import time
from datetime import datetime, timedelta
from pulp import *
from collections import defaultdict, deque

# Erstelle einen FileHandler und einen StreamHandler
file_handler = logging.FileHandler('logfile.log')
console_handler = logging.StreamHandler()
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(message)s', handlers=[file_handler, console_handler])

def count_crossings_via_variables(c_vars):
    crossings = 0
    for c_var in c_vars.values():
        if c_var.varValue == 1:
            crossings += 1
    return crossings

def solve_bipartite_minimization(graph_file):
    start_time = time()
    logging.info(f"Prozess für {graph_file} gestartet")
    base_name = os.path.basename(graph_file)
    new_base_name = base_name.replace('.gr', '.sol')
    new_base_name1 = base_name.replace('.gr', '.cros')
    output_file = os.path.join('mytests/solutions', new_base_name)
    output_file1 = os.path.join('mytests/crossings', new_base_name1)
    logging.info(f"Die Ausgabedatei wird {output_file} sein")

    edges = []
    with open(graph_file, "r") as file:
        for line in file:
            if line.startswith('c'):
                continue
            elif line.startswith('p'):
                parts = line.split()
                n0 = int(parts[2])
                n1 = int(parts[3])
                logging.info(f"Größen der Partitionen: A={n0}, B={n1}")
            else:
                x, y = map(int, line.split())
                edges.append((x, y))
    logging.info(f"{len(edges)} Kanten geladen.")

    prob = LpProblem("Minimize_Crossings", LpMinimize)
    y = {(i, j): LpVariable(f"y_{i}_{j}", 0, 1, cat='Binary') for i in range(n0 + 1, n0 + n1 + 1) for j in range(n0 + 1, n0 + n1 + 1) if i != j}
    c = {(i, j, k, l): LpVariable(f"c_{i}_{j}_{k}_{l}", 0, 1, cat='Binary') for (i, j) in edges for (k, l) in edges}
    logging.info("x, y und c geladen.")

    prob += lpSum(c.values())
    logging.info("Zielfunktion aufgestellt.")

    for (i, j) in edges:
        for (k, l) in edges:
            if k > i:
                if j > l:
                    prob += c[(i, j, k, l)] == y[(l, j)]
                elif l > j:
                    prob += c[(i, j, k, l)] == 1 - y[(j, l)]
    logging.info("Crossing Constraints aufgestellt.")

    def check_and_add_cycle_constraints(prob, y, n0, n1):
        graph = defaultdict(list)
        in_degree = defaultdict(int)
        nodes = range(n0 + 1, n0 + n1 + 1)
        for i in nodes:
            for j in nodes:
                if i != j:
                    y_ij = y.get((i, j))
                    if y_ij is not None:
                        if y_ij.varValue == 1:
                            graph[i].append(j)
                            in_degree[j] += 1
                        elif y_ij.varValue == 0:
                            graph[j].append(i)
                            in_degree[i] += 1

        def find_cycle(graph, nodes):
            visited = set()
            stack = []
            stack_set = set()

            def dfs(node):
                if node in stack_set:
                    return stack[stack.index(node):]  # Return the cycle
                stack.append(node)
                stack_set.add(node)
                visited.add(node)
                for neighbor in graph[node]:
                    result = dfs(neighbor)
                    if result:
                        return result
                stack.pop()
                stack_set.remove(node)
                return None

            for node in nodes:
                if node not in visited:
                    result = dfs(node)
                    if result:
                        return result
            return None

        cycle = find_cycle(graph, nodes)
        if cycle:
            logging.info(f"Zyklus entdeckt: {cycle}")
            y_values = {(cycle[i], cycle[j]): y[(cycle[i], cycle[j])].varValue for i in range(len(cycle)) for j in range(i + 1, len(cycle))}
            logging.info(f"y Werte im Zyklus: {y_values}")

            # Füge Constraints für die Kantenpaare im Zyklus hinzu
            for i in range(len(cycle)):
                for j in range(i + 1, len(cycle)):
                    prob += y[(cycle[i], cycle[j])] + y[(cycle[j], cycle[i])] == 1
                    logging.info(f"Constraint hinzugefügt für: {(cycle[i], cycle[j]), (cycle[j], cycle[i])}")
                    for k in range(j + 1, len(cycle)):
                        if (cycle[i], cycle[j]) in y and (cycle[j], cycle[k]) in y and (cycle[i], cycle[k]) in y:
                            prob += 0 <= y[(cycle[i], cycle[j])] + y[(cycle[j], cycle[k])] <= 1 + y[(cycle[i], cycle[k])]
                            #prob += y[(cycle[i], cycle[j])] + y[(cycle[i], cycle[k])] <= 1 + y[(cycle[j], cycle[k])]
                            logging.info(f"Constraint hinzugefügt für: {(cycle[i], cycle[j]), (cycle[j], cycle[k]), (cycle[i], cycle[k])}")

            return True
        return False

    while True:
        prob.solve()
        logging.info(f"Status der Lösung: {LpStatus[prob.status]}")
        if prob.status != LpStatusOptimal:
            logging.warning("Keine optimale Lösung gefunden.")
            break
        if not check_and_add_cycle_constraints(prob, y, n0, n1):
            break

    if prob.status == LpStatusOptimal:
        logging.info("Optimale Lösung gefunden. Ergebnisse werden gespeichert.")

        graph = defaultdict(list)
        in_degree = defaultdict(int)

        nodes = range(n0 + 1, n0 + n1 + 1)
        for i in nodes:
            for j in nodes:
                if i != j:
                    y_ij = y.get((i, j))
                    if y_ij is not None:
                        if y_ij.varValue == 1:
                            graph[i].append(j)
                            in_degree[j] += 1
                        elif y_ij.varValue == 0:
                            graph[j].append(i)
                            in_degree[i] += 1

        zero_in_degree_queue = deque([i for i in nodes if in_degree[i] == 0])
        sorted_b = []
        while zero_in_degree_queue:
            node = zero_in_degree_queue.popleft()
            sorted_b.append(node)
            for neighbor in graph[node]:
                in_degree[neighbor] -= 1
                if in_degree[neighbor] == 0:
                    zero_in_degree_queue.append(neighbor)

        cycle = [node for node in nodes if node not in sorted_b]
        print(cycle)
        a = count_crossings_via_variables(c)
        os.makedirs(os.path.dirname(output_file), exist_ok=True)
        with open(output_file, 'w') as f:
            for b in sorted_b:
                f.write(f"{b}\n")
                print(f"{b}")
        print("Crossings: ", a) 
        logging.info(f"Ergebnisse in {output_file} gespeichert")

        a = count_crossings_via_variables(c)

        os.makedirs(os.path.dirname(output_file1), exist_ok=True)
        with open(output_file1, 'w') as f:
            f.write(f"Crossings: {a}\n")     
        logging.info(f"Crossings: {a}, in {output_file1} gespeichert")

        end_time = time()
        elapsed_time = end_time - start_time
        elapsed_time_td = timedelta(seconds=elapsed_time)
        formatted_time = str(elapsed_time_td)
        formatted_time = formatted_time[:-3] 
        logging.info(f"Verstrichene Zeit: {formatted_time} (h:m:s.ms) \n")

def process_directory(directory_path):
    for filename in os.listdir(directory_path):
        if filename.endswith('.gr'):
            file_path = os.path.join(directory_path, filename)
            solve_bipartite_minimization(file_path)
            logging.info(f"Verarbeitung abgeschlossen für {file_path}")

#directory_path = 'githubtests/tiny_test_set/instances/'
#directory_path = 'mytests/instances/'
#process_directory(directory_path)

#test_file = 'githubtests/tiny_test_set/instances/complete_4_5.gr'
test_file = 'githubtests/medium_test_set/instances/11.gr'
#test_file = 'mytests/instances/1.gr'
solve_bipartite_minimization(test_file)
