import os
import logging
import time
from datetime import datetime, timedelta
from pulp import *
# Erstellen eines Graphen zur Bestimmung der Knotenreihenfolge
from collections import defaultdict, deque


# Erstelle einen FileHandler und einen StreamHandler
file_handler = logging.FileHandler('logfile.log')
console_handler = logging.StreamHandler()
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s', handlers=[file_handler, console_handler])

def count_crossings_via_variables(c_vars):
    crossings = 0
    for c_var in c_vars.values():
        if c_var.varValue == 1:
            crossings += 1
    return(crossings)

def detect_cycle(graph, node, visited, rec_stack, path):
    visited[node] = True
    rec_stack[node] = True
    path.append(node)

    for neighbor in graph[node]:
        if not visited[neighbor]:
            if detect_cycle(graph, neighbor, visited, rec_stack, path):
                return True
        elif rec_stack[neighbor]:
            path.append(neighbor)
            return True

    rec_stack[node] = False
    path.pop()
    return False

def has_cycle(graph, nodes):
    visited = {node: False for node in nodes}
    rec_stack = {node: False for node in nodes}
    path = []
    
    for node in nodes:
        if not visited[node]:
            if detect_cycle(graph, node, visited, rec_stack, path):
                cycle_start = path.index(path[-1])
                return True, path[cycle_start:]
    
    return False, []

def solve_bipartite_minimization(graph_file):
    start_time = time()
    logging.info(f"Prozess für {graph_file} gestartet")
    # Extrahiere den Basisnamen der Eingabedatei
    base_name = os.path.basename(graph_file)
    new_base_name = base_name.replace('.gr', '.sol')
    new_base_name1 = base_name.replace('.gr', '.cros')
    # Erstelle den Ausgabepfad
    output_file = os.path.join('mytests/solutions', new_base_name)
    output_file1 = os.path.join('mytests/crossings', new_base_name1)
    logging.info(f"Die Ausgabedatei wird {output_file} sein")

    edges = []
    with open(graph_file, "r") as file:
        for line in file:
            if line.startswith('c'):
                continue
            elif line.startswith('p'):
                parts = line.split()
                n0 = int(parts[2])  # Anzahl der Knoten in A
                n1 = int(parts[3])  # Anzahl der Knoten in B
                logging.info(f"Größen der Partitionen: A={n0}, B={n1}")
            else:
                x, y = map(int, line.split())
                edges.append((x, y))
    logging.info(f"{len(edges)} Kanten geladen.")

    prob = LpProblem("Minimize_Crossings", LpMinimize)

    # Boolesche Variablen für relative Positionen innerhalb jeder Partition
    y = {(i, j): LpVariable(f"y_{i}_{j}", 0, 1, cat='Binary') for i in range(n0 + 1, n0 + n1 + 1) for j in range(n0 + 1, n0 + n1 + 1) if i < j}
    c = {(i, j, k, l): LpVariable(f"c_{i}_{j}_{k}_{l}", 0, 1, cat='Binary') for (i, j) in edges for (k, l) in edges}# if i < j and k < l and i < k and j != l}
    logging.info("x, y und c geladen.")
    
    # Zielfunktion, die minimiert werden soll
    prob += lpSum(c.values())
    logging.info("Zielfunktion aufgestellt.")
    """
    edges.sort(key=lambda x: x[0])  # Sortieren der Kanten nach dem Startknoten
    for idx, (i, j) in enumerate(edges):
        for (k, l) in edges[idx + 1:]:
    """
    for (i, j) in edges:
        for (k, l) in edges:
            if k > i:  # Da die Liste sortiert ist, brauchen wir nur k > i zu prüfen
                #if (i, j, k, l) not in c:
                    #c[(i, j, k, l)] = LpVariable(f"c_{i}_{j}_{k}_{l}", 0, 1, cat='Binary')
                if j > l:
                    prob += c[(i, j, k, l)] == y[(l, j)]
                elif l > j:
                    prob += c[(i, j, k, l)] == 1 - y[(j, l)]
    logging.info("Crossing Constraints aufgestellt.")
    """
    counter = 0
    for (i, j) in y:
        for (k, l) in y:
            if j == l and (i, k) in y:
                prob += y[(i, k)] >= y[(i, j)] + y[(k, l)]
                counter += 1
            
            
            if j == k and (i, l) in y:
                prob += y[(i, l)] >= y[(i, j)] + y[(k, l)] - 1
                counter += 1
        
    logging.info(f"Es wurden {counter} Transitivitätsconstraints aufgestellt.")
    """
    """ 
    counter = 0
    for i in nodes:
        for j in nodes:
            for k in nodes:
                if i != j and j != k and i != k:
                    if (i, j) in y and (j, k) in y:
                        prob += y[(i, k)] >= y[(i, j)] + y[(j, k)] - 1
                        counter += 1
    logging.info(f"Es wurden {counter} Transitivitätsconstraints aufgestellt.")
    """

    prob.solve()
    logging.info(f"Status der Lösung: {LpStatus[prob.status]}")

    if prob.status == LpStatusOptimal:
        logging.info("Optimale Lösung gefunden. Ergebnisse werden gespeichert.")

        graph = defaultdict(list)
        in_degree = defaultdict(int)

        nodes = range(n0+1, n0+n1+1)
        for i in nodes:
            for j in nodes:
                if i != j:
                    y_ij = y.get((i, j))
                    
                    # Hinzufügen einer Kante basierend auf y_ij-Werten
                    if y_ij is not None:
                        if y_ij.value() == 1:
                            # i kommt vor j
                            graph[i].append(j)
                            in_degree[j] += 1
                        elif y_ij.value() == 0:
                            # j kommt vor i
                            graph[j].append(i)
                            in_degree[i] += 1
                            
        # Zyklusprüfung in den y-Variablen nach dem Lösen des ILP-Modells
        cycle_found, cycle_nodes = has_cycle(graph, nodes)
        if cycle_found:
            cycle_y_values = {(i, j): y[(i, j)].value() for i in cycle_nodes for j in cycle_nodes if (i, j) in y}
            logging.error(f"Zyklus gefunden in den y-Variablen: {cycle_nodes} mit y-Werten: {cycle_y_values}")
        
        # Topologische Sortierung
        zero_in_degree_queue = deque([i for i in nodes if in_degree[i] == 0])
        sorted_b = []
        while zero_in_degree_queue:
            node = zero_in_degree_queue.popleft()
            sorted_b.append(node)
            for neighbor in graph[node]:
                in_degree[neighbor] -= 1
                if in_degree[neighbor] == 0:
                    zero_in_degree_queue.append(neighbor)
        a = count_crossings_via_variables(c)
        # Ausgabe der sortierten Knoten
        os.makedirs(os.path.dirname(output_file), exist_ok=True)
        with open(output_file, 'w') as f:
            for b in sorted_b:
                f.write(f"{b}\n")
                print(f"{b}")
        print("Crossings: ", a) 
        logging.info(f"Ergebnisse in {output_file} gespeichert")
        # Ausgabe der sortierten Knoten

        a = count_crossings_via_variables(c)

        os.makedirs(os.path.dirname(output_file1), exist_ok=True)
        with open(output_file1, 'w') as f:
            f.write(f"Crossings: {a}\n")     
        logging.info(f"Crossings: {a}, in {output_file1} gespeichert")

        end_time = time()
        elapsed_time = end_time - start_time
        elapsed_time_td = timedelta(seconds=elapsed_time)
        formatted_time = str(elapsed_time_td)  # 
        formatted_time = formatted_time[:-3] 
        logging.info(f"Verstrichene Zeit: {formatted_time} (h:m:s.ms) \n")
    else:
        logging.warning("Keine optimale Lösung gefunden.")

        end_time = time()
        elapsed_time = end_time - start_time
        elapsed_time_td = timedelta(seconds=elapsed_time)
        formatted_time = str(elapsed_time_td)  #
        formatted_time = formatted_time[:-3] 
        logging.info(f"Verstrichene Zeit: {formatted_time} (h:m:s.ms) \n")


def process_directory(directory_path):
    # Durchlaufe alle Dateien im angegebenen Verzeichnis
    for filename in os.listdir(directory_path):
        if filename.endswith('.gr'):  # Überprüfen, ob die Datei eine .gr Datei ist
            file_path = os.path.join(directory_path, filename)
            solve_bipartite_minimization(file_path)  # Rufe die Verarbeitungsfunktion für jede .gr Datei auf
            logging.info(f"Verarbeitung abgeschlossen für {file_path}")

#directory_path = 'githubtests/tiny_test_set/instances/'
#directory_path = 'mytests/instances/'
#process_directory(directory_path)

#test_file = 'githubtests/tiny_test_set/instances/complete_4_5.gr'
test_file = 'githubtests/medium_test_set/instances/10.gr'
#test_file = 'mytests/instances/1.gr'
solve_bipartite_minimization(test_file)
