import os
import logging
import time
from datetime import datetime, timedelta
from pulp import *
# Erstellen eines Graphen zur Bestimmung der Knotenreihenfolge
from collections import defaultdict, deque


# Erstelle einen FileHandler und einen StreamHandler
file_handler = logging.FileHandler('len_cycles.log')
console_handler = logging.StreamHandler()
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s', handlers=[file_handler, console_handler])

def count_crossings_via_variables(c_vars):
    crossings = 0
    for c_var in c_vars.values():
        if c_var.varValue == 1:
            crossings += 1
    return(crossings)

def detect_cycle(graph, node, visited, rec_stack, path, cycles):
    visited[node] = True
    rec_stack[node] = True
    path.append(node)

    for neighbor in graph[node]:
        if not visited[neighbor]:
            detect_cycle(graph, neighbor, visited, rec_stack, path, cycles)
        elif rec_stack[neighbor]:
            cycle_start = path.index(neighbor)
            cycles.append(path[cycle_start:].copy())

    rec_stack[node] = False
    path.pop()

def find_all_cycles(graph, nodes):
    visited = {node: False for node in nodes}
    rec_stack = {node: False for node in nodes}
    path = []
    cycles = []

    for node in nodes:
        if not visited[node]:
            detect_cycle(graph, node, visited, rec_stack, path, cycles)

    # Entferne doppelte Zyklen
    unique_cycles = []
    seen = set()
    for cycle in cycles:
        cycle_tuple = tuple(sorted(cycle))
        if cycle_tuple not in seen:
            seen.add(cycle_tuple)
            unique_cycles.append(cycle)

    return unique_cycles

def solve_bipartite_minimization(graph_file):
    start_time = time()
    logging.info(f"Prozess für {graph_file} gestartet")
    # Extrahiere den Basisnamen der Eingabedatei
    base_name = os.path.basename(graph_file)
    new_base_name = base_name.replace('.gr', '.sol')
    new_base_name1 = base_name.replace('.gr', '.cros')
    # Erstelle den Ausgabepfad
    output_file = os.path.join('mytests/solutions', new_base_name)
    output_file1 = os.path.join('mytests/crossings', new_base_name1)

    edges = []
    with open(graph_file, "r") as file:
        for line in file:
            if line.startswith('c'):
                continue
            elif line.startswith('p'):
                parts = line.split()
                n0 = int(parts[2])  # Anzahl der Knoten in A
                n1 = int(parts[3])  # Anzahl der Knoten in B
                logging.info(f"Größen der Partitionen: A={n0}, B={n1}")
            else:
                x, y = map(int, line.split())
                edges.append((x, y))
    logging.info(f"{len(edges)} Kanten geladen.")

    prob = LpProblem("Minimize_Crossings", LpMinimize)

    # Boolesche Variablen für relative Positionen innerhalb jeder Partition
    y = {(i, j): LpVariable(f"y_{i}_{j}", 0, 1, cat='Binary') for i in range(n0 + 1, n0 + n1 + 1) for j in range(n0 + 1, n0 + n1 + 1) if i < j}
    c = {}
    
    # Zielfunktion, die minimiert werden soll
    prob += lpSum(c.values())


    for (i, j) in edges:
        for (k, l) in edges:
            if k > i:  # Da die Liste sortiert ist, brauchen wir nur k > i zu prüfen
                if j > l:
                    if (i, j, k, l) not in c:
                        c[(i, j, k, l)] = LpVariable(f"c_{i}_{j}_{k}_{l}", 0, 1, cat='Binary')
                    prob += c[(i, j, k, l)] == y[(l, j)]
                elif l > j:
                    if (i, j, k, l) not in c:
                        c[(i, j, k, l)] = LpVariable(f"c_{i}_{j}_{k}_{l}", 0, 1, cat='Binary')
                    prob += c[(i, j, k, l)] == 1 - y[(j, l)]

    prob.solve()
    logging.info(f"Status der Lösung: {LpStatus[prob.status]}")

    if prob.status == LpStatusOptimal:
        logging.info("Optimale Lösung gefunden. Ergebnisse werden gespeichert.")

        graph = defaultdict(list)
        in_degree = defaultdict(int)

        nodes = range(n0+1, n0+n1+1)
        for i in nodes:
            for j in nodes:
                if i != j:
                    y_ij = y.get((i, j))
                    
                    # Hinzufügen einer Kante basierend auf y_ij-Werten
                    if y_ij is not None:
                        if y_ij.value() == 1:
                            # i kommt vor j
                            graph[i].append(j)
                            in_degree[j] += 1
                        elif y_ij.value() == 0:
                            # j kommt vor i
                            graph[j].append(i)
                            in_degree[i] += 1
                            
        # Zyklusprüfung in den y-Variablen nach dem Lösen des ILP-Modells
        all_cycles = find_all_cycles(graph, nodes)
        logging.info(f"Alle gefundenen Zyklen: {len(all_cycles)} Stück.")

        end_time = time()
        elapsed_time = end_time - start_time
        elapsed_time_td = timedelta(seconds=elapsed_time)
        formatted_time = str(elapsed_time_td)  # 
        formatted_time = formatted_time[:-3] 
        logging.info(f"Verstrichene Zeit: {formatted_time} (h:m:s.ms) \n")
    else:
        logging.warning("Keine optimale Lösung gefunden.")

        end_time = time()
        elapsed_time = end_time - start_time
        elapsed_time_td = timedelta(seconds=elapsed_time)
        formatted_time = str(elapsed_time_td)  #
        formatted_time = formatted_time[:-3] 
        logging.info(f"Verstrichene Zeit: {formatted_time} (h:m:s.ms) \n")


def process_directory(directory_path):
    # Durchlaufe alle Dateien im angegebenen Verzeichnis
    for filename in os.listdir(directory_path):
        if filename.endswith('.gr'):  # Überprüfen, ob die Datei eine .gr Datei ist
            file_path = os.path.join(directory_path, filename)
            solve_bipartite_minimization(file_path)  # Rufe die Verarbeitungsfunktion für jede .gr Datei auf

directory_path = 'githubtests/tiny_test_set/instances/'
directory_path = 'githubtests/medium_test_set/instances/'
directory_path = 'mytests/instances/'
process_directory(directory_path)

#test_file = 'githubtests/tiny_test_set/instances/complete_4_5.gr'
#test_file = 'githubtests/medium_test_set/instances/10.gr'
#test_file = 'mytests/instances/1.gr'
#solve_bipartite_minimization(test_file)
