import os
import logging
from datetime import timedelta
import time
import argparse
from pulp import *
from collections import defaultdict, deque

# Konfiguriere Logging
file_handler = logging.FileHandler('logfile.log')
console_handler = logging.StreamHandler()
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s', handlers=[file_handler, console_handler])

def count_crossings_via_variables(c_vars):
    crossings = 0
    for c_var in c_vars.values():
        if c_var.varValue == 1:
            crossings += 1
    return crossings

def solve_bipartite_minimization(input_file, output_file):
    start_time = time()
    logging.info(f"Prozess für {input_file} gestartet")
    # Extrahiere den Basisnamen der Eingabedatei
    base_name = os.path.basename(input_file)
    # Erstelle den Ausgabepfad
    logging.info(f"Die Ausgabedatei wird {output_file} sein")

    edges = []
    with open(input_file, "r") as file:
        for line in file:
            if line.startswith('c'):
                continue
            elif line.startswith('p'):
                parts = line.split()
                n0 = int(parts[2])  # Anzahl der Knoten in A
                n1 = int(parts[3])  # Anzahl der Knoten in B
                logging.info(f"Größen der Partitionen: A={n0}, B={n1}")
            else:
                x, y = map(int, line.split())
                edges.append((x, y))
    logging.info(f"{len(edges)} Kanten geladen.")

    prob = LpProblem("Minimize_Crossings", LpMinimize)

    # Boolesche Variablen für relative Positionen innerhalb jeder Partition
    y = {(i, j): LpVariable(f"y_{i}_{j}", 0, 1, cat='Binary') for i in range(n0 + 1, n0 + n1 + 1) for j in range(n0 + 1, n0 + n1 + 1) if i < j}
    c = {(i, j, k, l): LpVariable(f"c_{i}_{j}_{k}_{l}", 0, 1, cat='Binary') for (i, j) in edges for (k, l) in edges if i < j and k < l and i < k and j != l}
    logging.info("x, y und c geladen.")
    
    # Zielfunktion, die minimiert werden soll
    prob += lpSum(c.values())
    logging.info("Zielfunktion aufgestellt.")

    edges.sort(key=lambda x: x[0])  # Sortieren der Kanten nach dem Startknoten
    for idx, (i, j) in enumerate(edges):
        for (k, l) in edges[idx + 1:]:
            if k > i:  # Da die Liste sortiert ist, brauchen wir nur k > i zu prüfen
                if (i, j, k, l) not in c:
                    c[(i, j, k, l)] = LpVariable(f"c_{i}_{j}_{k}_{l}", 0, 1, cat='Binary')
                if j > l:
                    prob += c[(i, j, k, l)] == y[(l, j)]
                elif l > j:
                    prob += c[(i, j, k, l)] == 1 - y[(j, l)]
    logging.info("Crossing Constraints aufgestellt.")

    prob.solve()
    logging.info(f"Status der Lösung: {LpStatus[prob.status]}")

    if prob.status == LpStatusOptimal:
        logging.info("Optimale Lösung gefunden. Ergebnisse werden gespeichert.")

        input = defaultdict(list)
        in_degree = defaultdict(int)

        nodes = range(n0+1, n0+n1+1)
        for i in nodes:
            for j in nodes:
                if i != j:
                    y_ij = y.get((i, j))
                    
                    # Hinzufügen einer Kante basierend auf y_ij-Werten
                    if y_ij is not None:
                        if y_ij.value() == 1:
                            # i kommt vor j
                            input[i].append(j)
                            in_degree[j] += 1
                        elif y_ij.value() == 0:
                            # j kommt vor i
                            input[j].append(i)
                            in_degree[i] += 1

        # Topologische Sortierung
        zero_in_degree_queue = deque([i for i in nodes if in_degree[i] == 0])
        sorted_b = []
        while zero_in_degree_queue:
            node = zero_in_degree_queue.popleft()
            sorted_b.append(node)
            for neighbor in input[node]:
                in_degree[neighbor] -= 1
                if in_degree[neighbor] == 0:
                    zero_in_degree_queue.append(neighbor)

        # Ausgabe der sortierten Knoten
        os.makedirs(os.path.dirname(output_file), exist_ok=True)
        with open(output_file, 'w') as f:
            for b in sorted_b:
                f.write(f"{b}\n")     
        logging.info(f"Ergebnisse in {output_file} gespeichert")
        # Ausgabe der sortierten Knoten

        a = count_crossings_via_variables(c)

        os.makedirs(os.path.dirname(output_file), exist_ok=True)
        with open(output_file, 'w') as f:
            for b in sorted_b:
                f.write(f"{b}\n")
                print(f"{b}")
        print("Crossings: ", a)

        end_time = time()
        elapsed_time = end_time - start_time
        elapsed_time_td = timedelta(seconds=elapsed_time)
        formatted_time = str(elapsed_time_td)  # '0:00:01.234000' für das obige Beispiel
        formatted_time = formatted_time[:-3] 
        logging.info(f"Verstrichene Zeit: {formatted_time} (h:m:s.ms) \n")
    else:
        logging.warning("Keine optimale Lösung gefunden.")

        end_time = time()
        elapsed_time = end_time - start_time
        elapsed_time_td = timedelta(seconds=elapsed_time)
        formatted_time = str(elapsed_time_td)  # '0:00:01.234000' für das obige Beispiel
        formatted_time = formatted_time[:-3] 
        logging.info(f"Verstrichene Zeit: {formatted_time} (h:m:s.ms) \n")


def main():
    parser = argparse.ArgumentParser(description="Solve Bipartite Minimization Problem")
    parser.add_argument('input_file', type=str, help="The path to the input file")
    parser.add_argument('output_file', type=str, help="The path to the output file")
    args = parser.parse_args()

    solve_bipartite_minimization(args.input_file, args.output_file)

if __name__ == "__main__":
    main()
