import math
import sys
import logging
from pulp import *
from collections import defaultdict, deque

def count_crossings_via_variables(c_vars):
    crossings = 0
    for c_var in c_vars.values():
        if c_var.varValue == 1:
            crossings += 1
    return crossings

def prioritize_cycles(cycles):
    cycles.sort(key=len, reverse=False)
    return cycles

def detect_cycle(graph, node, visited, rec_stack, path, cycles):
    visited[node] = True
    rec_stack[node] = True
    path.append(node)

    for neighbor in graph[node]:
        if not visited[neighbor]:
            detect_cycle(graph, neighbor, visited, rec_stack, path, cycles)
        elif rec_stack[neighbor]:
            cycle_start = path.index(neighbor)
            cycles.append(path[cycle_start:].copy())

    rec_stack[node] = False
    path.pop()

def find_all_cycles(graph, nodes):
    visited = {node: False for node in nodes}
    rec_stack = {node: False for node in nodes}
    path = []
    cycles = []

    for node in nodes:
        if not visited[node]:
            detect_cycle(graph, node, visited, rec_stack, path, cycles)

    unique_cycles = []
    seen = set()
    for cycle in cycles:
        cycle_tuple = tuple(sorted(cycle))
        if cycle_tuple not in seen:
            seen.add(cycle_tuple)
            unique_cycles.append(cycle)

    return unique_cycles

def add_cycle_constraints(prob, y, cycles, added_constraints):
    for cycle in cycles:
        cycle = list(dict.fromkeys(cycle))
        for i in range(len(cycle)):
            for j in range(i + 1, len(cycle)):
                constraint_1 = ((cycle[i], cycle[j]), (cycle[j], cycle[i]))
                if constraint_1 not in added_constraints:
                    prob += y[(cycle[i], cycle[j])] + y[(cycle[j], cycle[i])] == 1
                    added_constraints.add(constraint_1)
                    
                for k in range(j + 1, len(cycle)):
                    if (cycle[i], cycle[j]) in y and (cycle[j], cycle[k]) in y and (cycle[i], cycle[k]) in y:
                        constraint_2 = ((cycle[i], cycle[j]), (cycle[j], cycle[k]), (cycle[i], cycle[k]))
                        if constraint_2 not in added_constraints:
                            prob += 0 <= y[(cycle[i], cycle[j])] + y[(cycle[j], cycle[k])] <= 1 + y[(cycle[i], cycle[k])]
                            added_constraints.add(constraint_2)

def solve_bipartite_minimization(input_lines):
    edges = []
    degrees_b = defaultdict(int)
    neighbors_a = defaultdict(list)
    
    for line in input_lines:
        if line.startswith('c'):
            continue
        elif line.startswith('p'):
            parts = line.split()
            n0 = int(parts[2])
            n1 = int(parts[3])
        else:
            x, y = map(int, line.split())
            edges.append((x, y))
            degrees_b[y] += 1
            neighbors_a[y].append(x)
    print(neighbors_a)
    # Identify and sort nodes in partition B with degree 1
    degree_one_nodes = [node for node, degree in degrees_b.items() if degree == 1]
    degree_one_nodes.sort(key=lambda b: neighbors_a[b][0])

    # Remove degree one nodes from nodes list
    all_nodes = set(range(n0 + 1, n0 + n1 + 1))
    remaining_nodes = [node for node in all_nodes if node not in degree_one_nodes]

    # Remove edges that involve degree one nodes
    filtered_edges = [(i, j) for (i, j) in edges if j not in degree_one_nodes]

    # ILP for remaining nodes
    prob = LpProblem("Minimize_Crossings", LpMinimize)
    y = {(i, j): LpVariable(f"y_{i}_{j}", 0, 1, cat='Binary') for i in remaining_nodes for j in remaining_nodes if i != j}
    c = {(i, j, k, l): LpVariable(f"c_{i}_{j}_{k}_{l}", 0, 1, cat='Binary') for (i, j) in filtered_edges for (k, l) in filtered_edges}

    prob += lpSum(c.values())

    for (i, j) in filtered_edges:
            for (k, l) in filtered_edges:
                    if k > i:
                        if j > l:
                            prob += c[(i, j, k, l)] == y[(l, j)]
                        elif l > j:
                            prob += c[(i, j, k, l)] == 1 - y[(j, l)]

    added_constraints = set()
    iteration = 0
    max_cycle_length = 9

    while True:
        iteration += 1
        prob.solve(PULP_CBC_CMD(msg=0))

        if prob.status != LpStatusOptimal:
            break

        graph = defaultdict(list)
        in_degree = defaultdict(int)
        for i in remaining_nodes:
            for j in remaining_nodes:
                if i != j:
                    y_ij = y.get((i, j))
                    if y_ij is not None:
                        if y_ij.varValue == 1:
                            graph[i].append(j)
                            in_degree[j] += 1
                        elif y_ij.varValue == 0:
                            graph[j].append(i)
                            in_degree[i] += 1

        all_cycles = find_all_cycles(graph, remaining_nodes)
        
        if not all_cycles:
            break

        prioritized_cycles = prioritize_cycles(all_cycles)

        if len(prioritized_cycles[-1]) > max_cycle_length:
            k = min(int(math.floor(len(prioritized_cycles) / 4)), len(prioritized_cycles))
        else:
            k = len(prioritized_cycles)
        cycles_to_process = prioritized_cycles[:k]
        add_cycle_constraints(prob, y, cycles_to_process, added_constraints)

    # Collect the topologically sorted order of the remaining nodes
    if prob.status == LpStatusOptimal:
        zero_in_degree_queue = deque([i for i in remaining_nodes if in_degree[i] == 0])
        sorted_remaining_nodes = []
        while zero_in_degree_queue:
            node = zero_in_degree_queue.popleft()
            sorted_remaining_nodes.append(node)
            for neighbor in graph[node]:
                in_degree[neighbor] -= 1
                if in_degree[neighbor] == 0:
                    zero_in_degree_queue.append(neighbor)

        # Combine the sorted degree one nodes and remaining nodes using a new ILP
        combined_prob = LpProblem("Combine_Orders", LpMinimize)
        combined_y = {(i, j): LpVariable(f"combined_y_{i}_{j}", 0, 1, cat='Binary') for i in range(n0 + 1, n0 + n1 + 1)  for j in range(n0 + 1, n0 + n1 + 1)  if i != j}
        combined_c = {(i, j, k, l): LpVariable(f"combined_c_{i}_{j}_{k}_{l}", 0, 1, cat='Binary') for (i, j) in edges for (k, l) in edges}

        combined_prob += lpSum(combined_c.values())

        for (i, j) in edges:
            for (k, l) in edges:
                if k > i:
                    if j > l:
                        combined_prob += combined_c[(i, j, k, l)] == combined_y[(l, j)]
                    elif l > j:
                        combined_prob += combined_c[(i, j, k, l)] == 1 - combined_y[(j, l)]

        # Constraints for degree one nodes
        for i in range(len(degree_one_nodes) - 1):
            combined_prob += combined_y[(degree_one_nodes[i], degree_one_nodes[i + 1])] == 1

        # Constraints for remaining nodes
        for i in range(len(sorted_remaining_nodes) - 1):
            combined_prob += combined_y[(sorted_remaining_nodes[i], sorted_remaining_nodes[i + 1])] == 1



    added_constraints = set()
    max_cycle_length = 9

    while True:
        combined_prob.solve(PULP_CBC_CMD(msg=0))

        if combined_prob.status != LpStatusOptimal:
            break

        # Build the graph based on the current solution
        graph = defaultdict(list)
        in_degree = defaultdict(int)
        nodes = range(n0 + 1, n0 + n1 + 1)
        for i in nodes:
            for j in nodes:
                if i != j:
                    combined_y_ij = combined_y.get((i, j))
                    if combined_y_ij is not None:
                        if combined_y_ij.varValue == 1:
                            graph[i].append(j)
                            in_degree[j] += 1
                        elif combined_y_ij.varValue == 0:
                            graph[j].append(i)
                            in_degree[i] += 1

        # Detect all cycles in the graph
        all_cycles = find_all_cycles(graph, nodes)
        
        if not all_cycles:
            break

        # Prioritize the cycles by their length
        prioritized_cycles = prioritize_cycles(all_cycles)

        # Calculate k based on the current percentage
        # Check the length of the longest cycle and adjust k accordingly
        if len(prioritized_cycles[-1]) > max_cycle_length:
            k = min(int(math.floor(len(prioritized_cycles)/4)), len(prioritized_cycles))
        else:
            k = len(prioritized_cycles)
        
        # Take only the first k cycles from the prioritized list
        cycles_to_process = prioritized_cycles[:k]
        add_cycle_constraints(combined_prob, combined_y, cycles_to_process, added_constraints)

    if prob.status == LpStatusOptimal:
        # Topologically sort the nodes based on the in-degree
        zero_in_degree_queue = deque([i for i in nodes if in_degree[i] == 0])
        sorted_b = []
        while zero_in_degree_queue:
            node = zero_in_degree_queue.popleft()
            sorted_b.append(node)
            for neighbor in graph[node]:
                in_degree[neighbor] -= 1
                if in_degree[neighbor] == 0:
                    zero_in_degree_queue.append(neighbor)
        return sorted_b
    else:
        return None

def main():
    input_lines = sys.stdin.read().strip().split('\n')
    result = solve_bipartite_minimization(input_lines)
    
    if result:
        print("\n".join(map(str, result)))
    else:
        print("No result")
        
if __name__ == "__main__":
    main()
