import math
import os
import logging
from datetime import timedelta
import time
import argparse
from pulp import *
from collections import defaultdict, deque

# Konfiguriere Logging
file_handler = logging.FileHandler('logfile.log')
console_handler = logging.StreamHandler()
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s', handlers=[file_handler, console_handler])

def count_crossings_via_variables(c_vars):
    crossings = 0
    for c_var in c_vars.values():
        if c_var.varValue == 1:
            crossings += 1
    return crossings

def prioritize_cycles(cycles):
    # Beispiel: Zyklen nach Länge priorisieren
    cycles.sort(key=len, reverse=False)
    return cycles

def detect_cycle(graph, node, visited, rec_stack, path, cycles):
    visited[node] = True
    rec_stack[node] = True
    path.append(node)

    for neighbor in graph[node]:
        if not visited[neighbor]:
            detect_cycle(graph, neighbor, visited, rec_stack, path, cycles)
        elif rec_stack[neighbor]:
            cycle_start = path.index(neighbor)
            cycles.append(path[cycle_start:].copy())

    rec_stack[node] = False
    path.pop()

def find_all_cycles(graph, nodes):
    visited = {node: False for node in nodes}
    rec_stack = {node: False for node in nodes}
    path = []
    cycles = []

    for node in nodes:
        if not visited[node]:
            detect_cycle(graph, node, visited, rec_stack, path, cycles)

    # Entferne doppelte Zyklen
    unique_cycles = []
    seen = set()
    for cycle in cycles:
        cycle_tuple = tuple(sorted(cycle))
        if cycle_tuple not in seen:
            seen.add(cycle_tuple)
            unique_cycles.append(cycle)

    return unique_cycles

def add_cycle_constraints(prob, y, cycles, added_constraints):
    for cycle in cycles:
        cycle = list(dict.fromkeys(cycle))
        for i in range(len(cycle)):
            for j in range(i + 1, len(cycle)):
                constraint_1 = ((cycle[i], cycle[j]), (cycle[j], cycle[i]))
                if constraint_1 not in added_constraints:
                    # einer von beiden ist wahr
                    prob += y[(cycle[i], cycle[j])] + y[(cycle[j], cycle[i])] == 1
                    added_constraints.add(constraint_1)
                    
                for k in range(j + 1, len(cycle)):
                    if (cycle[i], cycle[j]) in y and (cycle[j], cycle[k]) in y and (cycle[i], cycle[k]) in y:
                        constraint_2 = ((cycle[i], cycle[j]), (cycle[j], cycle[k]), (cycle[i], cycle[k]))
                        if constraint_2 not in added_constraints:
                            #i < j und j < l -> i< k reihenfolgen constraint transitivität
                            prob += 0 <= y[(cycle[i], cycle[j])] + y[(cycle[j], cycle[k])] <= 1 + y[(cycle[i], cycle[k])]
                            added_constraints.add(constraint_2)
        logging.info(f"Constraint hinzugefügt für: {cycle}")
    logging.info(f"Constraint hinzugefügt für alle bestehenden cycles.")

def solve_bipartite_minimization(input_file, output_file):
    start_time = time()
    logging.info(f"Prozess für {input_file} gestartet")
    # Extrahiere den Basisnamen der Eingabedatei
    base_name = os.path.basename(input_file)
    # Erstelle den Ausgabepfad
    logging.info(f"Die Ausgabedatei wird {output_file} sein")

    edges = []
    with open(input_file, "r") as file:
        for line in file:
            if line.startswith('c'):
                continue
            elif line.startswith('p'):
                parts = line.split()
                n0 = int(parts[2])
                n1 = int(parts[3])
                logging.info(f"Größen der Partitionen: A={n0}, B={n1}")
            else:
                x, y = map(int, line.split())
                edges.append((x, y))
    logging.info(f"{len(edges)} Kanten geladen.")

    prob = LpProblem("Minimize_Crossings", LpMinimize)
    y = {(i, j): LpVariable(f"y_{i}_{j}", 0, 1, cat='Binary') for i in range(n0 + 1, n0 + n1 + 1) for j in range(n0 + 1, n0 + n1 + 1) if i != j}
    c = {(i, j, k, l): LpVariable(f"c_{i}_{j}_{k}_{l}", 0, 1, cat='Binary') for (i, j) in edges for (k, l) in edges}
    logging.info("x, y und c geladen.")

    prob += lpSum(c.values())
    logging.info("Zielfunktion aufgestellt.")

    for (i, j) in edges:
        for (k, l) in edges:
            if k > i:
                if j > l:
                    prob += c[(i, j, k, l)] == y[(l, j)]
                elif l > j:
                    prob += c[(i, j, k, l)] == y[(l, j)]
    logging.info("Crossing Constraints aufgestellt.")

    added_constraints = set()
    iteration = 0
    max_cycle_length = 9

    while True:
        iteration += 1
        prob.solve()
        logging.info(f"Status der Lösung: {LpStatus[prob.status]}")
        if prob.status != LpStatusOptimal:
            logging.warning("Keine optimale Lösung gefunden.")
            break

        graph = defaultdict(list)
        in_degree = defaultdict(int)
        nodes = range(n0 + 1, n0 + n1 + 1)
        for i in nodes:
            for j in nodes:
                if i != j:
                    y_ij = y.get((i, j))
                    if y_ij is not None:
                        if y_ij.varValue == 1:
                            graph[i].append(j)
                            in_degree[j] += 1
                        elif y_ij.varValue == 0:
                            graph[j].append(i)
                            in_degree[i] += 1

        all_cycles = find_all_cycles(graph, nodes)
        logging.info(f"Alle gefundenen Zyklen: {len(all_cycles)} Stück.")
        
        if not all_cycles:
            break

        prioritized_cycles = prioritize_cycles(all_cycles)
        logging.info(f"Priorisierte Zyklen aufgestellt.")

        # Berechne k basierend auf dem aktuellen Prozentsatz
        # Überprüfe die Länge des längsten Zyklus und passe k entsprechend an
        if len(prioritized_cycles[-1]) > max_cycle_length:
            k = min(int(math.floor(len(prioritized_cycles)/4)), len(prioritized_cycles))
        else:
            k = len(prioritized_cycles)
        # Nehme nur die ersten k Zyklen aus der priorisierten Liste
        cycles_to_process = prioritized_cycles[:k]
        logging.info(f"Zyklen zur Verarbeitung in dieser Iteration: {cycles_to_process}")
        logging.info(f"Anzahl der Zyklen zur Verarbeitung in dieser Iteration: {len(cycles_to_process)}")
        add_cycle_constraints(prob, y, cycles_to_process, added_constraints)

    if prob.status == LpStatusOptimal:
        logging.info("Optimale Lösung gefunden. Ergebnisse werden gespeichert.")

        zero_in_degree_queue = deque([i for i in nodes if in_degree[i] == 0])
        sorted_b = []
        while zero_in_degree_queue:
            node = zero_in_degree_queue.popleft()
            sorted_b.append(node)
            for neighbor in graph[node]:
                in_degree[neighbor] -= 1
                if in_degree[neighbor] == 0:
                    zero_in_degree_queue.append(neighbor)

        a = count_crossings_via_variables(c)
        os.makedirs(os.path.dirname(output_file), exist_ok=True)
        with open(output_file, 'w') as f:
            for b in sorted_b:
                f.write(f"{b}\n")
                print(f"{b}")
        print("Crossings: ", a) 
        logging.info(f"Ergebnisse in {output_file} gespeichert")

        end_time = time()
        elapsed_time = end_time - start_time
        elapsed_time_td = timedelta(seconds=elapsed_time)
        formatted_time = str(elapsed_time_td)
        formatted_time = formatted_time[:-3] 
        logging.info(f"Verstrichene Zeit: {formatted_time} (h:m:s.ms) \n")
    else:
        logging.warning("Keine optimale Lösung gefunden.")
        end_time = time()
        elapsed_time = end_time - start_time
        elapsed_time_td = timedelta(seconds=elapsed_time)
        formatted_time = str(elapsed_time_td)
        formatted_time = formatted_time[:-3] 
        logging.info(f"Verstrichene Zeit: {formatted_time} (h:m:s.ms) \n")

def main():
    parser = argparse.ArgumentParser(description="Solve Bipartite Minimization Problem")
    parser.add_argument('input_file', type=str, nargs='?', help="The path to the input file")
    parser.add_argument('output_file', type=str, nargs='?', help="The path to the output file")
    args = parser.parse_args()

    if not args.input_file:
        print("Error: Es wurde keine Eingabedatei angegeben.")
        return

    input_file = args.input_file
    output_file = args.output_file if args.output_file else 'default_output.txt'

    logging.info(f"Prozess für {input_file} gestartet")
    logging.info(f"Die Ausgabedatei wird {output_file} sein")

    solve_bipartite_minimization(input_file, output_file)

if __name__ == "__main__":
    main()
